<?php

declare(strict_types=1);

/*
 * This file is part of the TYPO3 CMS project.
 *
 * It is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License, either version 2
 * of the License, or any later version.
 *
 * For the full copyright and license information, please read the
 * LICENSE.txt file that was distributed with this source code.
 *
 * The TYPO3 project - inspiring people to share!
 */

namespace ITSHofmann\LowLevel\Command;

use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;
use TYPO3\CMS\Backend\Command\ProgressListener\ReferenceIndexProgressListener;
use TYPO3\CMS\Core\Core\Bootstrap;
use TYPO3\CMS\Core\Core\Environment;
use TYPO3\CMS\Core\Database\ConnectionPool;
use TYPO3\CMS\Core\Database\ReferenceIndex;
use TYPO3\CMS\Core\Resource\ResourceStorage;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Core\Resource\StorageRepository;



/**
 * Finds files which are referenced by TYPO3 but not found in the file system
 */
class RemoveDeletedContent extends Command
{

    /**
     * Configure the command by defining the name, options and arguments
     */
    public function configure()
    {
        $this
            ->setDescription('RemoveDeletedContent.')
            
            ->addOption(
                'dry-run',
                null,
                InputOption::VALUE_NONE,
                'If this option is set, the references will not be removed, but just the output which files would be deleted are shown'
            );
            
    }

    /**
     * Executes the command to
     * - optionally update the reference index (to have clean data)
     * - find data in sys_refindex (softrefs and regular references) where the actual file does not exist (anymore)
     * - remove these files if --dry-run is not set (not possible for refindexes)
     *
     * @param InputInterface $input
     * @param OutputInterface $output
     * @return int
     */
    protected function execute(InputInterface $input, OutputInterface $output)
    {
        // Make sure the _cli_ user is loaded
        Bootstrap::initializeBackendAuthentication();

        $io = new SymfonyStyle($input, $output);
        $io->title($this->getDescription());        

        $dryRun = $input->hasOption('dry-run') && $input->getOption('dry-run') != false ? true : false;

        
        // Find missing soft references (cannot be updated / deleted)
        $deletedContent = $this->findDeletedContent($io,$dryRun);
        if ($deletedContent > 0) {
            $io->note('Found ' . $deletedContent . ' deleted content.');            
        }

        if ($deletedContent == 0 ) {
            $io->success('Nothing to do, deleted content found.');
        }
        return 0;
    }

    
    /**
     *
     * @return array an array of records within deletedContent
     */
    protected function findDeletedContent(SymfonyStyle $io,$dryRun): int
    {
        $contentToDelete = 0;
        // Select all files in the reference table
        /** @var  ConnectionPool $connectionPool */
        $connectionPool = GeneralUtility::makeInstance(ConnectionPool::class);
        $qBRef = $connectionPool
            ->getQueryBuilderForTable('tt_content');
        $qBRef            
            ->getRestrictions()
            ->removeAll();
        $count = $qBRef ->count('uid')
            ->from('tt_content')
            ->where(
                $qBRef->expr()->eq(
                    'deleted',$qBRef->createNamedParameter(1, \PDO::PARAM_INT
                )
            )
            )->execute()->fetchOne();
                
        $contentToDelete = $count;
        if (!$dryRun) {
            $qBRef ->delete('tt_content')
            ->where(
                $qBRef->expr()->eq('deleted',$qBRef->createNamedParameter(1, \PDO::PARAM_INT))
            )->execute();
        }   
        return $contentToDelete;
    }


   

    
}
