<?php

declare(strict_types=1);

/*
 * This file is part of the TYPO3 CMS project.
 *
 * It is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License, either version 2
 * of the License, or any later version.
 *
 * For the full copyright and license information, please read the
 * LICENSE.txt file that was distributed with this source code.
 *
 * The TYPO3 project - inspiring people to share!
 */

namespace ITSHofmann\LowLevel\Command;

use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;
use TYPO3\CMS\Backend\Command\ProgressListener\ReferenceIndexProgressListener;
use TYPO3\CMS\Core\Core\Bootstrap;
use TYPO3\CMS\Core\Core\Environment;
use TYPO3\CMS\Core\Database\ConnectionPool;
use TYPO3\CMS\Core\Database\ReferenceIndex;
use TYPO3\CMS\Core\Resource\ResourceStorage;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Core\Resource\StorageRepository;



/**
 * Finds files which are referenced by TYPO3 but not found in the file system
 */
class UnUsedFilesCommand extends Command
{

    /**
     * Configure the command by defining the name, options and arguments
     */
    public function configure()
    {
        $this
            ->setDescription('Delecte all files records with out references.')
            ->setHelp('
Assumptions:
- a perfect integrity of the reference index table (always update the reference index table before using this tool!)

Files may be missing for these reasons (except software bugs):
- someone manually deleted the file inside fileadmin/ or another user maintained folder. If the reference was a soft reference (opposite to a DataHandler managed file relation from "group" type fields), technically it is not an error although it might be a mistake that someone did so.
- someone manually deleted the file inside the uploads/ folder (typically containing managed files) which is an error since no user interaction should take place there.

Manual repair suggestions (using --dry-run):
- Managed files: You might be able to locate the file and re-insert it in the correct location. However, no automatic fix can do that for you.


If the option "--dry-run" is not set, all managed files (TCA/FlexForm attachments) will silently remove the reference
from the record since the file is missing. For this reason you might prefer a manual approach instead.
All soft references with missing files require manual fix if you consider it an error.

If you want to get more detailed information, use the --verbose option.')
            ->addOption(
                'dry-run',
                null,
                InputOption::VALUE_NONE,
                'If this option is set, the references will not be removed, but just the output which files would be deleted are shown'
            )
            ->addArgument(
                'exclude',
                InputArgument::OPTIONAL,
                'Exclude files from folder e.g. cleanup:deleteunusedfiles --dry-run  2:/Partner,1:/downloads'
            )
            ->addOption(
                'update-refindex',
                null,
                InputOption::VALUE_NONE,
                'Setting this option automatically updates the reference index and does not ask on command line. Alternatively, use -n to avoid the interactive mode'
            );
    }

    /**
     * Executes the command to
     * - optionally update the reference index (to have clean data)
     * - find data in sys_refindex (softrefs and regular references) where the actual file does not exist (anymore)
     * - remove these files if --dry-run is not set (not possible for refindexes)
     *
     * @param InputInterface $input
     * @param OutputInterface $output
     * @return int
     */
    protected function execute(InputInterface $input, OutputInterface $output)
    {
        // Make sure the _cli_ user is loaded
        Bootstrap::initializeBackendAuthentication();

        $io = new SymfonyStyle($input, $output);
        $io->title($this->getDescription());
        $io->note('Found ' .' non-used files.');
        $excludes = [];
        if ($input->getArgument('exclude')) {
            $excludes = explode(',',$input->getArgument('exclude'));
        }
        $io->write('exclude[s] '.implode(',',$excludes));

        $dryRun = $input->hasOption('dry-run') && $input->getOption('dry-run') != false ? true : false;

        // Update the reference index
        $this->updateReferenceIndex($input, $io);

        // Find missing soft references (cannot be updated / deleted)
        $sys_files_withoutRef = $this->findSysFilesWithoutRef($io,$excludes,$dryRun);
        if (count($sys_files_withoutRef)) {
            $io->note('Found ' . count($sys_files_withoutRef) . ' non-used files.');
            $io->write('exclude[s] '.implode(',',$excludes));
            if ($dryRun) {
                $io->success('All above files would be deleted.');
            } else {
                $io->success('All above files are no deleted.');
            }
            
        }

        if (!count($sys_files_withoutRef) ) {
            $io->success('Nothing to do, no missing files found. Everything is in place.');
        }
        return 0;
    }

    /**
     * Function to update the reference index
     * - if the option --update-refindex is set, do it
     * - otherwise, if in interactive mode (not having -n set), ask the user
     * - otherwise assume everything is fine
     *
     * @param InputInterface $input holds information about entered parameters
     * @param SymfonyStyle $io necessary for outputting information
     */
    protected function updateReferenceIndex(InputInterface $input, SymfonyStyle $io)
    {
        // Check for reference index to update
        $io->note('Finding missing files referenced by TYPO3 requires a clean reference index (sys_refindex)');
        if ($input->hasOption('update-refindex') && $input->getOption('update-refindex')) {
            $updateReferenceIndex = true;
        } elseif ($input->isInteractive()) {
            $updateReferenceIndex = $io->confirm('Should the reference index be updated right now?', false);
        } else {
            $updateReferenceIndex = false;
        }

        // Update the reference index
        if ($updateReferenceIndex) {
            $progressListener = GeneralUtility::makeInstance(ReferenceIndexProgressListener::class);
            $progressListener->initialize($io);
            $referenceIndex = GeneralUtility::makeInstance(ReferenceIndex::class);
            $referenceIndex->updateIndex(false, $progressListener);
        } else {
            $io->writeln('Reference index is assumed to be up to date, continuing.');
        }
    }

    /**
     * Find file references that points to non-existing files in system
     * Fix methods: API in \TYPO3\CMS\Core\Database\ReferenceIndex that allows to
     * change the value of a reference (or remove it)
     *
     * @return array an array of records within sys_refindex
     */
    protected function findSysFilesWithoutRef(SymfonyStyle $io,$excludes,$dryRun): array
    {
        $unusedFiles = [];
        // Select all files in the reference table
        /** @var  ConnectionPool $connectionPool */
        $connectionPool = GeneralUtility::makeInstance(ConnectionPool::class);
        $qBRef = $connectionPool
            ->getQueryBuilderForTable('sys_file_reference');
        $qBRef ->count('uid')
            ->from('sys_file_reference')
            ->where(
                $qBRef->expr()->eq(
                    'table_local',$qBRef->createNamedParameter('sys_file', \PDO::PARAM_STR)
                )
            )->andWhere(
                $qBRef->expr()->eq(
                    'uid_local',$qBRef->createNamedParameter(0, \PDO::PARAM_INT, ':sysfile_uid')
                )
            );


        $queryBuilder = $connectionPool
            ->getQueryBuilderForTable('sys_file');

        $result = $queryBuilder
            ->select('*')
            ->from('sys_file')
            ->where(
                $queryBuilder->expr()->gt('storage', $queryBuilder->createNamedParameter(0, \PDO::PARAM_INT))
            )
            ->execute();

        // Traverse the references and check if the files exists
        while ($record = $result->fetchAssociative()) {
        //while ($record = $result->fetch()) {
            $qBRef->setParameter('sysfile_uid',$record['uid']);
            $count = $qBRef->execute()
                ->fetchOne();
            if ($count == 0) {
                $fileName = $this->getFileNameWithoutAnchor($record['identifier']);
                $excludeThis = false;
                foreach ($excludes as $exclude) {
                    [$storage, $folder] = explode(':', $exclude);
                    if ($storage == $record['storage']) {
                        $pos = strpos($record['identifier'], $folder);
                        if ($pos !== false) {
                            if ($pos == 0) {
                                $excludeThis = true;
                            }
                        }
                    }
                }
                if (!$excludeThis) {
                    $unusedFiles[$record['uid']] = $fileName;
                    if (!$dryRun) {
                        $this->removeSysFile($record,$dryRun,$io);
                    }
                    $io->writeln('Deleting file "' . $fileName . '":'. $record['uid'].' '.$record['storage'].' ' .$record['identifier']);
                }
            }
        }
        return $unusedFiles;
    }


    /**
     * Remove a possible anchor like 'my-path/file.pdf#page15'
     *
     * @param string $fileName a filename as found in sys_refindex.ref_string
     * @return string the filename but leaving everything behind #page15 behind
     */
    protected function getFileNameWithoutAnchor(string $fileName): string
    {
        if (strpos($fileName, '#') !== false) {
            [$fileName] = explode('#', $fileName);
        }
        return $fileName;
    }

    /**
     * Removes all references in the sys_file_reference where files were not found
     *
     * @param array $missingManagedFiles Contains the records of sys_refindex which need to be updated
     * @param bool $dryRun if set, the references are just displayed, but not removed
     * @param SymfonyStyle $io the IO object for output
     */
    protected function removeSysFile(array $record, bool $dryRun, SymfonyStyle $io)
    {
        GeneralUtility::makeInstance(ConnectionPool::class)->getConnectionForTable('sys_file')
            ->delete(
                'sys_file',
                [ 'uid' =>  $record['uid'] ] // where
            );
        $fileName = $this->getFileNameWithoutAnchor($record['identifier']);
        $fileName = ltrim($fileName,'/');
        $storageRepository = GeneralUtility::makeInstance(StorageRepository::class);
        /** @var StorageRepository $storageRepository */
        /** @var ResourceStorage $storage */
        if (intval(($record['storage'])>0)){
            $storage = $storageRepository->findByUid($record['storage']);
            $path = $storage->getPublicUrl($storage->getRootLevelFolder());
            if (@is_file(Environment::getPublicPath().'/' . $path . $fileName)) {
                if (@file_exists(Environment::getPublicPath().'/' . $path . $fileName)) {
                    unlink(Environment::getPublicPath().'/' . $path . $fileName);
                }
            }
        } else {
            if ( @is_file(Environment::getPublicPath() . $fileName)) {
                if ( !@file_exists(Environment::getPublicPath() . $fileName)) {
                //    unlink(Environment::getPublicPath().'/' . $fileName);
                }
            }
        }
    }
}
